/**********************************************************************************
 * Source file for the implementation of the algorithm for
 * computing rearrangement distance of all permutations of size n.
 * Author : Gustavo Rodrigues Galvao - May 2010.
**********************************************************************************/


#include "commons.h"

tThread_args *newThreadArgs(int n, int sign, void *func, int assimetryLevel){
	tThread_args *args;

	args = (tThread_args *) malloc(sizeof(tThread_args));
	args->n = n;
	args->sign = sign;
	args->assimetryLevel = assimetryLevel;
	args->loadVector = (void *)func;

	return args;
}

void initializeAuxiliaryData(int n, int sign, int **temp, int **copy, tPermutation **permutation){
	int *pElements;

	(*temp) = (int*) malloc(sizeof(int)*n);
	(*copy) = (int*) malloc(sizeof(int)*n);
	pElements = (int*) malloc(sizeof(int)*n);
	(*permutation) = newPermutation(pElements,n,sign);
}

void freeAuxiliaryData(int *temp, int *copy, tPermutation *permutation){
	free(temp);
	free(copy);
	freePermutation(permutation);
}

void initializeThreadData(int n, int sign, int **temp, int **copy, tPermutation **permutation, unsigned int **vector){
	initializeAuxiliaryData(n, sign, temp, copy, permutation);
	(*vector) = (unsigned int*) malloc(sizeof(unsigned int)*(n*n*n*n*n));
}

void freeThreadData(int *temp, int *copy, tPermutation *permutation, unsigned int *vector){
	freeAuxiliaryData(temp, copy, permutation);
	free(vector);
}

void initializeCommonData(int n, int sign, unsigned int size){
	int *temp, *copy;
	unsigned int current;
	tPermutation *permutation;

	slices = (int*) malloc(sizeof(int)*500);
	initializeVector(slices,500);
	queue = (unsigned int*) malloc(sizeof(unsigned int)*size);
	hash = (char*) malloc(sizeof(char)*size);
	initializeHash(hash,size);

	initializeAuxiliaryData(n, sign, &temp, &copy, &permutation);

	getIdentityPermutation(permutation);
	current = rankPermutation(permutation, temp, copy);
	queue[0] = current;
	indexFirst = 0;
	indexLast = 1%(size);
	hash[current] = (char) 0;

	freeAuxiliaryData(temp, copy, permutation);
}

void freeCommonData(){
	free(queue);
	free(hash);
	free(slices);
}

void processNextPermutation(tPermutation *permutation, unsigned int* current, int *d){
	*current = queue[indexFirst];
	indexFirst = indexFirst + 1;
	*d = (int) hash[*current];
	unrankPermutation(permutation, *current);
}

void waitIfAnotherSlice(int d, int previousD){
	int aux;

	if(d > previousD){
		pthread_mutex_lock(&mutexSlices);
		slices[previousD]--;
		slices[d]++;
		
		if(slices[currentSlice] == 0) {
			currentSlice++;		
		}

		aux = currentSlice;
		pthread_mutex_unlock(&mutexSlices);

		while(d > aux){
			sleep(1);
			pthread_mutex_lock(&mutexSlices);
			aux = currentSlice;
			pthread_mutex_unlock(&mutexSlices);
		}
	}
}

void insertInQueue(int d, unsigned int next, unsigned int current){
	if(hash[next] == ((char) 255)){
		queue[indexLast] = next;
		indexLast = indexLast + 1;
		hash[next] = (char) (d+1);
	}
}

void printPermutations(int n, int sign, unsigned int size){
	int d;
	int *pElements;
	unsigned int i, current;
	tPermutation *permutation;

	pElements = (int*) malloc(sizeof(int)*n);
	permutation = newPermutation(pElements,n,sign);

	for(i=0; i<size; i++) {
		current = queue[i];
		d = (int) hash[current];
		unrankPermutation(permutation,current);
		printOutput(permutation,d);
	}

	freePermutation(permutation);
}

void printDistances(unsigned int size){
	unsigned int i;

	for(i=0; i<size; i++) {
		printf("%c",hash[i]);
	}
}

int queueHasPermutations(unsigned int size){
	if(indexLast < size){
		return 1;
	} else {
		return 0;
	}
}

void unloadVector(int comb, int d, unsigned int *vector, unsigned int current){
	int k;

	for(k = 0; k < comb; k++){
		insertInQueue(d, vector[k], current);
	}
}

void *thread(void *arg){
	int *temp, *copy;
	int d,previousD,n,comb,sign,assimetryLevel;
	tPermutation *permutation;
	unsigned int size, *vector, current;
	tThread_args *args;

   	args = (tThread_args*) arg;
	n = (int) args->n;
	sign = (int) args->sign;
	assimetryLevel = (int) args->assimetryLevel;
	size = factorial(n);
	if(sign == 1){
		size = size*base2[n];
	}
	initializeThreadData(n, sign, &temp, &copy, &permutation, &vector);

	pthread_mutex_lock(&mutexSlices);
	d = currentSlice;
	slices[d]++;
	pthread_mutex_unlock(&mutexSlices);

	pthread_mutex_lock(&mutexCommonData);
	while(queueHasPermutations(size)) {
		previousD = d;
		processNextPermutation(permutation, &current, &d);
		pthread_mutex_unlock(&mutexCommonData);
		waitIfAnotherSlice(d,previousD);
		args->loadVector(n, &comb, temp, copy, permutation, vector, assimetryLevel);
		pthread_mutex_lock(&mutexCommonData);
		unloadVector(comb, d, vector, current);
	}
	pthread_mutex_unlock(&mutexCommonData);

	pthread_mutex_lock(&mutexSlices);
	slices[d]--;
	pthread_mutex_unlock(&mutexSlices);

	freeThreadData(temp, copy, permutation, vector);
}

void runThreads(int n, int sign, int num_threads, int assimetryLevel, void *func){
	int t;
	pthread_t *threads;
	pthread_attr_t attr;
	tThread_args *args;

	threads = (pthread_t*) malloc(sizeof(pthread_t)*num_threads);
   	pthread_attr_init(&attr);
   	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  	pthread_mutex_init(&mutexCommonData, NULL);
	pthread_mutex_init(&mutexSlices, NULL);
	args = newThreadArgs(n, sign, func, assimetryLevel);
	currentSlice = 0;

	for(t=0; t < num_threads; t++){
       		pthread_create(&threads[t], NULL, thread, (void *) args);
		sleep(5);
     	}

	for(t=0; t < num_threads; t++) {
      		pthread_join(threads[t], NULL);
	}

	free(threads);
	free(args);
}

void generateAllPermutations(int n, int sign, int num_threads, int assimetryLevel, int printASCII, void *loadVector){
	unsigned int size;

	size = factorial(n);
	if(sign == 1){
		base2 = (unsigned int*) malloc(sizeof(unsigned int)*(n+1));
		getBase2(base2,n+1);
		size = size*base2[n];
	}

	initializeCommonData(n, sign, size);
	runThreads(n, sign, num_threads, assimetryLevel, loadVector);

	if(printASCII){
		printPermutations(n, sign, size);
	} else {
		printDistances(size);
	}

	freeCommonData();

	if(sign == 1){
		free(base2);
	}
}

unsigned int nextReversal(tPermutation *permutation, int *temp, int *copy, int i, int j){
	unsigned int id;

	applyReversal(permutation,temp,i,j);
	id = rankPermutation(permutation, temp, copy);
	applyReversal(permutation,temp,i,j);

	return id;
}

unsigned int nextTransposition(tPermutation *permutation, int *temp, int *copy, int i, int j, int k){
	unsigned int id;

	applyTransposition(permutation,temp,i,j,k);
	id = rankPermutation(permutation, temp, copy);
	applyTransposition(permutation,temp,i,k-j+i,k);

	return id;
}

unsigned int nextTransreversalA(tPermutation *permutation, int *temp, int *copy, int i, int j, int k){
	unsigned int id;

	applyTransreversalA(permutation,temp,i,j,k);
	id = rankPermutation(permutation, temp, copy);
	applyTransreversalB(permutation,temp,i,k-j+i,k);

	return id;
}

unsigned int nextTransreversalB(tPermutation *permutation, int *temp, int *copy, int i, int j, int k){
	unsigned int id;

	applyTransreversalB(permutation,temp,i,j,k);
	id = rankPermutation(permutation, temp, copy);
	applyTransreversalA(permutation,temp,i,k-j+i,k);

	return id;
}

unsigned int nextSignedTransreversalA(tPermutation *permutation, int *temp, int *copy, int i, int j, int k){
	unsigned int id;

	applySignedTransreversalA(permutation,temp,i,j,k);
	id = rankPermutation(permutation, temp, copy);
	applySignedTransreversalB(permutation,temp,i,k-j+i,k);

	return id;
}

unsigned int nextSignedTransreversalB(tPermutation *permutation, int *temp, int *copy, int i, int j, int k){
	unsigned int id;

	applySignedTransreversalB(permutation,temp,i,j,k);
	id = rankPermutation(permutation, temp, copy);
	applySignedTransreversalA(permutation,temp,i,k-j+i,k);

	return id;
}

unsigned int nextBlockInterchange(tPermutation *permutation, int *temp, int *copy, int i, int j, int k, int l){
	unsigned int id;

	applyBlockInterchange(permutation,temp,i,j,k,l);
	id = rankPermutation(permutation, temp, copy);
	applyBlockInterchange(permutation,temp,i,i+l-k,i+l-j,l);

	return id;
}

unsigned int nextSignedReversal(tPermutation *permutation, int *temp, int *copy, int i, int j){
	unsigned int id;

	applySignedReversal(permutation,temp,i,j);
	id = rankPermutation(permutation, temp, copy);
	applySignedReversal(permutation,temp,i,j);

	return id;
}

unsigned int nextCenteredReversal(tPermutation *permutation, int *temp, int *copy, int i, int j){
	unsigned int id;

	applyCenteredReversal(permutation,temp,i,j);
	id = rankPermutation(permutation, temp, copy);
	applyCenteredReversal(permutation,temp,i,j);
	
	return id;
}

unsigned int nextSignedCenteredReversal(tPermutation *permutation, int *temp, int *copy, int i, int j){
	unsigned int id;

	applySignedCenteredReversal(permutation,temp,i,j);
	id = rankPermutation(permutation, temp, copy);
	applySignedCenteredReversal(permutation,temp,i,j);

	return id;
}

unsigned int nextSwap(tPermutation *permutation, int *temp, int *copy, int i, int j){
	unsigned int id;

	applySwap(permutation,temp,i,j);
	id = rankPermutation(permutation, temp, copy);
	applySwap(permutation,temp,i,j);

	return id;
}

